/*
 * fastsid.c - MOS6581 (SID) emulation.
 *
 * Written by
 *  Teemu Rantanen <tvr@cs.hut.fi>
 *  Michael Schwendt <sidplay@geocities.com>
 *  Ettore Perazzoli <ettore@comm2000.it>
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 *
 */

/**** Changes by Juan Linietsky for inclusing in legasynth ***

	Changed processing function to work with signed integers
 	Several wrappings into sid struct
  	Separated init functions
**********/


//#include "vice.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/*
#include "log.h"
#include "machine.h"
#include "maincpu.h"
#include "resources.h"
#include "sid.h"
#include "sound.h"
#include "snapshot.h"
*/

#ifdef __cplusplus
extern "C" {
#endif

#include "sid_types.h"
//#include "utils.h"
#include "fixpoint.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

/* use wavetables (sampled waveforms) */
#define WAVETABLES

/* ADSR state */
#define SID_ENV_ATTACK   0
#define SID_ENV_DECAY    1
#define SID_ENV_SUSTAIN  2
#define SID_ENV_RELEASE  3
#define SID_ENV_IDLE     4

#ifndef WAVETABLES
/* Current waveform */
#define TESTWAVE          0
#define PULSEWAVE         1
#define SAWTOOTHWAVE      2
#define TRIANGLEWAVE      3
#define NOISEWAVE         4
#define NOWAVE            5
#define RINGWAVE          6
#define PULSETRIANGLEWAVE 7
#define PULSESAWTOOTHWAVE 8
#endif

/* noise magic */
#define NSHIFT(v, n) \
    (((v) << (n))    \
    | ((((v) >> (23 - (n))) ^ (v >> (18 - (n)))) & ((1 << (n)) - 1)))

#define NVALUE(v)                                   \
    (noiseLSB[v & 0xff] | noiseMID[(v >> 8) & 0xff] \
    | noiseMSB[(v >> 16) & 0xff])

#define NSEED 0x7ffff8


#include "wave6581.h"
#include "wave8580.h"

typedef struct _Filter {

	vreal_t Dy, ResDy;
	BYTE Type;
	BYTE CurType;
	WORD Value;
	float RefFreql;

} Filter;
/* needed data for one voice */
typedef struct voice_s
{

    Filter *filterp;
    struct sound_s	*s;
    struct voice_s	*vprev;
    struct voice_s	*vnext;
    int			 nr;
    int nosustain;
    /* counter value */
    DWORD		 f;
    /* counter step / sample */
    DWORD		 fs;
#ifdef WAVETABLES
    /* do we have noise enabled? */
    BYTE		 noise;
#else
    /* waveform that we use */
    BYTE		 fm;
    /* pulse threshold compared to the 32-bit counter */
    DWORD		 pw;
#endif

    /* 31-bit adsr counter */
    DWORD		 adsr;
    /* adsr counter step / sample */
    SDWORD		 adsrs;
    /* adsr sustain level compared to the 31-bit counter */
    DWORD		 adsrz;

    /* does this voice use hard sync? */
    BYTE		 sync;
    /* does this voice use filter? */
    BYTE		 filter;
    /* does this structure need updating before next sample? */
    BYTE		 update;
    /* did we do multiple gate flips after last calculated sample? */
    BYTE		 gateflip;

    /* ADSR mode */
    BYTE		 adsrm;
    /* 4-bit attack value */
    BYTE		 attack;
    /* 4-bit decay value */
    BYTE		 decay;
    /* 4-bit sustain value */
    BYTE		 sustain;
    /* 4-bit release value */
    BYTE		 release;

    /* pointer to registers of this voice */
    BYTE		*d;

    /* noise shift register. Note! rv may be 0 to 15 shifts 'behind' the
       real noise shift register value. Remaining shifts are done when
       it is referenced */
    DWORD		 rv;
#ifdef WAVETABLES
    /* pointer to wavetable data */
    WORD		*wt;
    /* 32-bit offset to add to the counter before referencing the wavetable.
       This is used on combined waveforms, when other waveforms are combined
       with pulse */
    DWORD		 wtpf;
    /* length of wavetable (actually number of shifts needed for 32-bit
       counter) */
    DWORD		 wtl;
    /* kludge for ring modulation. Set wtr[1] = 0x7fff if ring modulation is
       used */
    WORD		 wtr[2];
#endif

    signed char		filtIO;
    vreal_t		filtLow, filtRef;
} voice_t;



/* needed data for SID */
struct sound_s
{

	//i'm placing this here, global variables suck

    Filter filter;

    CLOCK clk; //clock!
    /* number of voices */
    voice_t		 v[3];
    /* SID registers */
    BYTE		 d[32];
    /* is voice 3 enabled? */
    BYTE		 has3;
    /* 4-bit volume value */
    BYTE		 vol;

    /* ADSR counter step values for each adsr values */
    SDWORD		 adrs[16];
    /* sustain values compared to 31-bit ADSR counter */
    DWORD		 sz[16];

    /* internal constant used for sample rate dependent calculations */
    DWORD		 speed1;

    /* does this structure need updating before next sample? */
    BYTE		 update;
#ifdef WAVETABLES
    /* do we have a new sid or an old one? */
    BYTE		 newsid;
#endif
    /* constants needed to implement write-only register reads */
    BYTE		 laststore;
    BYTE		 laststorebit;
    CLOCK		 laststoreclk;
    /* do we want to use filters? */
    BYTE		 emulatefilter;
};

typedef struct sound_s sound_t;

BYTE fastsid_read(sound_t *psid, ADDRESS addr);
void fastsid_store(sound_t *psid, ADDRESS addr, BYTE byte);
void fastsid_avoid_sustain(sound_t *psid,int p_voice);
void fastsid_close(sound_t *psid);
int fastsid_init(sound_t *psid, int speed, int cycles_per_sec);
void init_filter(int freq);
int fastsid_calculate_samples(sound_t *psid, SWORD *pbuf, int nr,int interleave, int *delta_t);
void init_wavetables();


#ifdef __cplusplus
};
#endif
